/******************************************************************************/
/*                                                                            */
/*                 X r d S e c P r o t o c o l s s l . c c                    */
/*                                                                            */
/* (c) 2007 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

//       $Id: XrdSecProtocolssl.cc,v 1.2 2007/12/10 04:12:05 abh Exp $

const char *XrdSecProtocolsslCVSID = "$Id: XrdSecProtocolssl.cc,v 1.0 2000/05/01 04:12:05 apeters Exp $";

#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <iostream>
#include <stdlib.h>
#include <strings.h>
#include <grp.h>
#include <pwd.h>

#include <openssl/crypto.h>
#include <openssl/x509v3.h>
#include <openssl/ssl.h>
#include <openssl/pem.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <fcntl.h>

#include "XrdNet/XrdNetDNS.hh"
#include "XrdOuc/XrdOucErrInfo.hh"
#include "XrdOuc/XrdOucHash.hh"
#include "XrdOuc/XrdOucString.hh"
#include "XrdOuc/XrdOucTrace.hh"
#include "XrdOuc/XrdOucTokenizer.hh"
#include "XrdSys/XrdSysPthread.hh"
#include "XrdSys/XrdSysLogger.hh"
#include "XrdSec/XrdSecInterface.hh"
#include "XrdSec/XrdSecTLayer.hh"
#include "XrdSecssl/XrdSecProtocolsslTrace.hh"
#include "libsslGridSite/grst_verifycallback.h"
#include "libsslGridSite/gridsite.h"

// fix for SSL 098 stuff and g++ 

#ifdef R__SSL_GE_098
#undef PEM_read_SSL_SESSION
#undef PEM_write_SSL_SESSION

#define PEM_read_SSL_SESSION(fp,x,cb,u) (SSL_SESSION *)PEM_ASN1_read( (void *(*)(void **, const unsigned char **, long int))d2i_SSL_SESSION,PEM_STRING_SSL_SESSION,fp,(void **)x,cb,u)

#define PEM_write_SSL_SESSION(fp,x) PEM_ASN1_write((int (*)(void*, unsigned char**))i2d_SSL_SESSION, PEM_STRING_SSL_SESSION,fp, (char *)x,  NULL,NULL,0,NULL,NULL)
//#else
//#ifdef __macos__
//#undef PEM_read_SSL_SESSION
//#undef PEM_write_SSL_SESSION

//#define PEM_read_SSL_SESSION(fp,x,cb,u) (SSL_SESSION *)PEM_ASN1_read( (char *(*)(...))d2i_SSL_SESSION,PEM_STRING_SSL_SESSION,fp,(char **)x  ,cb,u)

//#define PEM_write_SSL_SESSION(fp,x) PEM_ASN1_write((int (*)(...))i2d_SSL_SESSION, PEM_STRING_SSL_SESSION,fp, (char *)x, NULL,NULL,0,NULL,NULL)

//#endif
#endif

#define l2n(l,c)        (*((c)++)=(unsigned char)(((l)>>24)&0xff), \
                         *((c)++)=(unsigned char)(((l)>>16)&0xff), \
                         *((c)++)=(unsigned char)(((l)>> 8)&0xff), \
                         *((c)++)=(unsigned char)(((l)    )&0xff))

static XrdOucTrace        *SSLxTrace=0;

class XrdSecProtocolssl;

#define MAX_SESSION_ID_ATTEMPTS 10

/******************************************************************************/
/*              X r d S e c P r o t o c o l s s l C l a s s                   */
/******************************************************************************/

class XrdSecsslSessionLock {
private:
static  XrdSysMutex sessionmutex;
  int sessionfd;
  
public:
  XrdSecsslSessionLock() {sessionfd=0;}
  bool SoftLock() { sessionmutex.Lock();}
  bool HardLock(const char* path) {sessionfd = open(path,O_RDWR); if ( (sessionfd>0) && (!flock(sessionfd,LOCK_EX)))return true;}
  ~XrdSecsslSessionLock() {if (sessionfd>0) {flock(sessionfd,LOCK_UN);close(sessionfd);};sessionmutex.UnLock();}
  bool SoftUnLock() {sessionmutex.UnLock();}
  bool HardUnLock() {if (sessionfd>0) {flock(sessionfd,LOCK_UN);close(sessionfd);sessionfd=0;}}
};



class XrdSecProtocolssl : public XrdSecTLayer
{
public:
  friend class XrdSecProtocolDummy; // Avoid stupid gcc warnings about destructor

  XrdSecProtocolssl(const char* hostname, const struct sockaddr  *ipaddr) : XrdSecTLayer("ssl",XrdSecTLayer::isServer) {
    credBuff    = 0;
    ssl         = 0;
    Entity.name = 0;
    Entity.grps = 0;
    host        = hostname;
    if (ipaddr)
      Entity.host = (XrdNetDNS::getHostName((sockaddr&)*ipaddr));
    else 
      Entity.host = strdup("");
    sprintf(proxyBuff,"");
    client_cert=0;
    server_cert=0;
    ssl = 0 ;
  }
  
  
  virtual void   secClient(int theFD, XrdOucErrInfo      *einfo);
  virtual void   secServer(int theFD, XrdOucErrInfo      *einfo=0);

  virtual void              Delete() {delete this;}


  static int GenerateSession(const SSL* ssl, unsigned char *id, unsigned int *id_len);
  static int NewSession(SSL* ssl, SSL_SESSION *pNew);
  static int GetSession(SSL* ssl, SSL_SESSION *pNew);

  static char*              SessionIdContext ;
  static char*              sslcadir; 
  static char*              sslvomsdir; 
  static char*              sslkeyfile;
  static char*              sslcertfile;
  static char*              sslproxyexportdir;
  static int                debug;
  static time_t             sslsessionlifetime;
  static bool               isServer;
  static bool               forwardProxy;
  static bool               allowSessions;
  static X509_STORE*        store;  
  static X509_LOOKUP*       lookup;
  static int                verifydepth;
  static int                verifyindex;
  int                       sessionfd;
  X509*    client_cert; 
  X509*    server_cert;
  XrdOucString              host;

  // for error logging and tracing
  static XrdSysLogger       Logger;
  static XrdSysError        eDest;
  static time_t             storeLoadTime;
  
  typedef struct {
    int verbose_mode;
    int verify_depth;
    int always_continue;
  } sslverify_t;
  
  char proxyBuff[8192];
  static SSL_CTX* ctx;

  XrdSysMutex SSLMutex;
private:
  
  ~XrdSecProtocolssl() {
    if (credBuff)    free(credBuff);
    if (Entity.name) free(Entity.name);
    if (Entity.grps) free(Entity.grps);
    if (Entity.host) free(Entity.host);
    SSLMutex.Lock();
    if (ssl) SSL_free(ssl);
    if (client_cert) X509_free(client_cert);
    if (server_cert) X509_free(server_cert);
    SSLMutex.UnLock();    
  } 

static int Fatal(XrdOucErrInfo *erp, const char* msg, int rc);
  
  
  struct sockaddr           hostaddr;      // Client-side only
  char                     *credBuff;      // Credentials buffer (server)
  int                       Step;          // Indicates step in authentication
  
  int sd;
  int listen_sd;
  struct sockaddr_in sa_serv;
  struct sockaddr_in sa_cli;
  SSL*     ssl;
};

/******************************************************************************/
/*             C l i e n t   O r i e n t e d   F u n c t i o n s              */
/******************************************************************************/

int XrdSecProtocolssl::Fatal(XrdOucErrInfo *erp, const char *msg, int rc)
{
  const char *msgv[8];
  int k, i = 0;
  
  msgv[i++] = "Secssl: ";    //0
  msgv[i++] = msg;            //1

  if (erp) erp->setErrInfo(rc, msgv, i);
  else {for (k = 0; k < i; k++) cerr <<msgv[k];
  cerr <<endl;
  }
  
  return -1;
}

/******************************************************************************/
/*                        g e t C r e d e n t i a l s                         */
/******************************************************************************/


void   
XrdSecProtocolssl::secClient(int theFD, XrdOucErrInfo      *error) {
  struct passwd *pEnt;
  struct group  *pGrp;
  char Buff[1024], *Bp;
  int Blen;
  
  EPNAME("secClient");

  char *nossl = getenv("XrdSecNoSSL");
  if (nossl) {
    error->setErrInfo(ENOENT,"SSL is disabled by force");
    return ;
  }

  error->setErrInfo(0,"");
  SSLMutex.Lock();  
  
  int err;
  int sd;
  char*    str;
  char     buf [4096];
  SSL_METHOD *meth;
  SSL_SESSION *session=0;

  SSL_load_error_strings();  
  SSLeay_add_ssl_algorithms();
  meth = (SSL_METHOD*) SSLv3_client_method();

  ERR_load_crypto_strings();

  XrdOucString sslsessionfile="";
  XrdOucString sslsessionid="";
  
  sslsessionfile = "/tmp/xssl_";
  sslsessionid += (int)geteuid();
  sslsessionid += ":";
  sslsessionid += host.c_str();
  sslsessionfile += sslsessionid;

  XrdSecsslSessionLock sessionlock;
  sessionlock.SoftLock();

  if (allowSessions) {
    struct stat sessionstat;

    if (!stat(sslsessionfile.c_str(),&sessionstat)) {
      // session exists ... I try to read it
      if (sessionlock.HardLock(sslsessionfile.c_str())) {
	FILE* fp = fopen(sslsessionfile.c_str(), "r");
	if (fp) {
	  {
	    session = PEM_read_SSL_SESSION(fp, NULL, NULL, NULL);
	    fclose(fp);
	    if (session) {
	      
	      DEBUG("info: ("<<__FUNCTION__<<") Session loaded from " << sslsessionfile.c_str());
	      char session_id[1024];
	      for (int i=0; i< session->session_id_length; i++) {
		sprintf(session_id+(i*2),"%02x",session->session_id[i]);
	      }
	    
	      if (session->cipher) {
		DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Cipher: " << session->cipher->name  << " Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
	      } else {
		unsigned char buf[5],*p;
                unsigned long l;
		
                p=buf;
                l=session->cipher_id;
                l2n(l,p);
                if ((session->ssl_version>>8) == SSL3_VERSION_MAJOR)
		  session->cipher=meth->get_cipher_by_char(&(buf[2]));
                else
		  session->cipher=meth->get_cipher_by_char(&(buf[1]));
                if (session->cipher == NULL) {
		  SSL_SESSION_free(session);
		  delete session;
		  session = NULL;		  
		} else {
		  DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Cipher: " << session->cipher->name  << " Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
		//		DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
		}
	      }
	    } else {
	    DEBUG("info: ("<<__FUNCTION__<<") Session load failed from " << sslsessionfile.c_str());
	    ERR_print_errors_fp(stderr);
	    }
	  }
	}
      }
    } else {
    }
  }

  ctx = SSL_CTX_new (meth);

  SSL_CTX_set_options(ctx,  SSL_OP_ALL | SSL_OP_NO_SSLv2);

  if (session) {
    SSL_CTX_add_session(ctx,session);
  }

  if (!ctx) {
    Fatal(error,"Cannot do SSL_CTX_new",-1);
    exit(2);
  }
  
  if (SSL_CTX_use_certificate_chain_file(ctx, sslcertfile) <= 0) {
    ERR_print_errors_fp(stderr);
    exit(3);
  }
  
  if (SSL_CTX_use_PrivateKey_file(ctx, sslkeyfile, SSL_FILETYPE_PEM) <= 0) {
    ERR_print_errors_fp(stderr);
    exit(4);
  }
  
  if (!SSL_CTX_check_private_key(ctx)) {
    fprintf(stderr,"Error: (%s) Private key does not match the certificate public key\n",__FUNCTION__);
    exit(5);
  } else {
    DEBUG("Private key check passed ...");
  }
  
  SSL_CTX_load_verify_locations(ctx, NULL,sslcadir);
  SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,  GRST_callback_SSLVerify_wrapper);
  //  SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE,  GRST_callback_SSLVerify_wrapper);
  SSL_CTX_set_cert_verify_callback(ctx, GRST_verify_cert_wrapper, (void *) NULL);

  grst_cadir   = sslcadir;
  grst_vomsdir = sslvomsdir;

  //  ctx->verify_depth=verifydepth;
  grst_depth=verifydepth;
  SSL_CTX_set_verify_depth(ctx, verifydepth);
  
  ssl = SSL_new (ctx);            

  SSL_set_purpose(ssl,X509_PURPOSE_ANY);

  if (session) {
    SSL_set_session(ssl, session);
  }

  sessionlock.SoftUnLock();
  sessionlock.HardUnLock();

  //  if (ssl && (ssl->kssl_ctx = kssl_ctx_new()) != NULL) {
  //    const char* host = "pcitsmd01.cern.ch";
  //    kssl_ctx_setstring(con->kssl_ctx,KSSL_SERVER, host);
  //  }

  if (!ssl) {
    Fatal(error,"Cannot do SSL_new",-1);
    exit(6);
  }

  SSL_set_fd (ssl, theFD);

  err = SSL_connect (ssl);

  if (err!=1) {
    // we want to see the error message from the server side
    ERR_print_errors_fp(stderr);
    if (theFD>=0) {close(theFD);theFD=-1;}
    SSLMutex.UnLock();  
    return;
  }

  session = SSL_get_session(ssl);
  
  if (session) {
    char session_id[1024];
    for (int i=0; i< session->session_id_length; i++) {
      sprintf(session_id+(i*2),"%02x",session->session_id[i]);
    }
    
    if (session->cipher) {
      DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Cipher: " << session->cipher->name  << " Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
    } else {
      DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
    }
    // write out the session
    FILE* fp = fopen((const char*)(sslsessionfile.c_str()),"w+");
    if (fp) {
      SSL_CTX_set_session_id_context(XrdSecProtocolssl::ctx,(const unsigned char*) XrdSecProtocolssl::SessionIdContext,
				     strlen(XrdSecProtocolssl::SessionIdContext));
      
      PEM_write_SSL_SESSION(fp, session);
      fclose(fp);
      DEBUG("info: ("<<__FUNCTION__<<") Session stored to " << sslsessionfile.c_str());
    }
  }
  
  /* Get the cipher - opt */
  
  TRACE(Authen,"SSL connection uses cipher: "<<SSL_get_cipher (ssl));
  
  /* Get server's certificate (note: beware of dynamic allocation) - opt */
  
  server_cert = SSL_get_peer_certificate (ssl);       
  XrdOucString rdn;
  
  str = X509_NAME_oneline (X509_get_subject_name (server_cert),0,0);
  rdn = str;
  TRACE(Authen,"Server certificate subject:\t" << str);
  OPENSSL_free (str);
  
  str = X509_NAME_oneline (X509_get_issuer_name  (server_cert),0,0);
  TRACE(Authen,"Server certificate  issuer: \t" << str);
  OPENSSL_free (str);
  
  /* We could do all sorts of certificate verification stuff here before
     deallocating the certificate. */
  
  X509_free (server_cert);
  server_cert = 0;

  if (forwardProxy) {
    if (!strcmp(sslkeyfile,sslcertfile)) {
      // this is a cert & key in one file atleast ... looks like proxy
      int fd = open(sslkeyfile,O_RDONLY);
      if (fd>=0) {
	int nread = read(fd,proxyBuff, sizeof(proxyBuff));
	if (nread>=0) {
	  TRACE(Authen,"Uploading my Proxy ...\n");
	  err = SSL_write(ssl, proxyBuff,nread);
	  if (err!= nread) {
	    Fatal(error,"Cannot forward proxy",-1);
	    if (theFD>=0) {close(theFD); theFD=-1;}
	    SSLMutex.UnLock();  
	    return;
	  }
	  char ok[16];
	  err = SSL_read(ssl,ok, 3);
	  if (err != 3) {
	    Fatal(error,"Didn't receive OK",-1);
	    if (theFD>=0) {close(theFD); theFD=-1;}
	    SSLMutex.UnLock();  
	    return;
	  } 
	} else {
	  close(fd);
	  Fatal(error,"Cannot read proxy file to forward",-1);
	  if (theFD>=0) {close(theFD);theFD=-1;}
	  SSLMutex.UnLock();  
	  return;
	}
      } else {
	Fatal(error,"Cannot read proxy file to forward",-1);
	if (theFD>=0) {close(theFD);theFD=-1;}
	SSLMutex.UnLock();  
	return;
      }
      close(fd);
    }
  }


  if (theFD>=0) {close(theFD);theFD=-1;}  
  SSLMutex.UnLock();  
  return;
}

/******************************************************************************/
/*               S e r v e r   O r i e n t e d   M e t h o d s                */
/******************************************************************************/


void MyGRSTerrorLogFunc (char *lfile, int lline, int llevel, char *fmt, ...) {
  EPNAME("grst");
  va_list args;
  char fullmessage[4096];

  va_start(args, fmt);
  vsprintf(fullmessage,fmt,args);
  va_end(args);

  // just remove linefeeds
  XrdOucString sfullmessage = fullmessage;
  sfullmessage.replace("\n","");


  if (llevel <= GRST_LOG_WARNING) {
    TRACE(Authen," ("<< lfile << ":" << lline <<"): " << sfullmessage);    
  } else if (llevel <= GRST_LOG_INFO) {
    TRACE(Authen, " ("<< lfile << ":" << lline <<"): " << sfullmessage);    
  } else {
    DEBUG(" ("<< lfile << ":" << lline <<"): " << sfullmessage);
  }
}

void   
XrdSecProtocolssl::secServer(int theFD, XrdOucErrInfo      *error) {
  int err;

  struct sockaddr_in sa_serv;
  struct sockaddr_in sa_cli;
  socklen_t client_len=0;
  char*    str;
  char     buf [4096];

  EPNAME("secServer");

  XrdSecsslSessionLock sessionlock; 

  // check if we should reload the store
  if ((time(NULL)-storeLoadTime) > 3600) {
    if (store) {
      TRACE(Authen,"Reloading X509 Store from " << sslcadir);
      X509_STORE_free(store);
      store = SSL_X509_STORE_create(NULL, sslcadir);
      X509_STORE_set_flags(XrdSecProtocolssl::store,0);
      storeLoadTime = time(NULL);
    } 
  }
  SSLMutex.Lock();  
  SSL_CTX_set_session_cache_mode(XrdSecProtocolssl::ctx, SSL_SESS_CACHE_BOTH | SSL_SESS_CACHE_NO_AUTO_CLEAR );

  ssl = SSL_new (ctx);                           

  SSL_set_purpose(ssl,X509_PURPOSE_ANY);

  SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_BOTH | SSL_SESS_CACHE_NO_AUTO_CLEAR );
  SSL_CTX_set_session_id_context(ctx,(const unsigned char*) XrdSecProtocolssl::SessionIdContext,
  				 strlen(XrdSecProtocolssl::SessionIdContext));  

  TRACE(Authen,"Info: ("<<__FUNCTION__<<") Session Cache has size: " <<SSL_CTX_sess_get_cache_size(ctx));


  if (!ssl) {
    fprintf(stderr,"Error: (%s) failed to create context\n",__FUNCTION__);
    TRACE(Authen,"Error: ("<<__FUNCTION__<<") failed to create context");
    exit(5);
  }

  SSL_set_fd (ssl, theFD);

  // the accept run's without mutex
  err = SSL_accept (ssl);                        

  if (err!=1) {
    long verifyresult = SSL_get_verify_result(ssl);
    if (verifyresult != X509_V_OK) {
      Fatal(error,X509_verify_cert_error_string(verifyresult),verifyresult);
      TRACE(Authen,"Error: ("<<__FUNCTION__<<") failed SSL_accept ");
    } else {
      printf("Received error %d/%d\n",err,SSL_get_error(ssl,err));
      Fatal(error,"do SSL_accept",-1);
      unsigned long lerr;
      while (lerr=ERR_get_error()) {TRACE(Authen,"SSL Queue error: err=" << lerr << " msg=" <<
					  ERR_error_string(lerr, NULL));Fatal(error,ERR_error_string(lerr,NULL),-1);}
      //      Fatal(error,"Failed SSL accept",-1);
    }
    SSLMutex.UnLock();
    if (theFD>=0) {close(theFD); theFD=-1;}
    return;
  }
  
  SSL_SESSION* session = SSL_get_session(ssl);

  if (session) {
    char session_id[1024];
    for (int i=0; i< session->session_id_length; i++) {
      sprintf(session_id+(i*2),"%02x",session->session_id[i]);
    }
    
    if (session->cipher) {
      DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Cipher: " << session->cipher->name  << "Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
    } else {
      DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
    }
    DEBUG("Info: ("<<__FUNCTION__<<") cache items             : " << SSL_CTX_sess_number(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") client connects         : " << SSL_CTX_sess_connect(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") client renegotiates     : " << SSL_CTX_sess_connect_renegotiate(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") client connect finished : " << SSL_CTX_sess_connect_good(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") server accepts          : " << SSL_CTX_sess_accept(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") server renegotiates     : " << SSL_CTX_sess_accept_renegotiate(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") server accepts finished : " << SSL_CTX_sess_accept_good(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") session cache hits      : " << SSL_CTX_sess_hits(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") session cache misses    : " << SSL_CTX_sess_misses(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") session cache timeouts  : " << SSL_CTX_sess_timeouts(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") callback cache hits     : " << SSL_CTX_sess_cb_hits(ctx));
    DEBUG("Info: ("<<__FUNCTION__<<") cache full overflows    : " << SSL_CTX_sess_cache_full(ctx) << " allowed: " << SSL_CTX_sess_get_cache_size(ctx));
    
  }
  /* get the grst_chain  */
  GRSTx509Chain *grst_chain = (GRSTx509Chain*) SSL_get_app_data(ssl);
  
  char* vomsstring=0;
  XrdOucString vomsroles="";
  XrdOucString clientdn="";
  
  if (grst_chain) {
    GRST_print_ssl_creds((void*) grst_chain);
    char* vr = GRST_get_voms_roles_and_free((void*) grst_chain);
    if (vr) {
      vomsroles = vr;
      free(vr);
    }
  }
  
  TRACE(Authen,"Authenticated with VOMS roles: "<<vomsroles);
  
  long verifyresult = SSL_get_verify_result(ssl);
  
  TRACE(Authen,"Verify result is = "<<verifyresult);
  
  /* Get the cipher - opt */
  
  DEBUG("SSL connection uses cipher " << SSL_get_cipher(ssl));
  
  /* Get client's certificate (note: beware of dynamic allocation) - opt */
  
  client_cert = SSL_get_peer_certificate (ssl);


  if (client_cert != NULL) {
    str = X509_NAME_oneline (X509_get_subject_name (client_cert), 0, 0);
    if (str) {
      TRACE(Authen,"client certificate subject: "<< str);
      clientdn = str;
      OPENSSL_free (str);
    } else {
      TRACE(Authen,"client certificate subject: none");
    }
    
    str = X509_NAME_oneline (X509_get_issuer_name  (client_cert), 0, 0);
    
    if (str) {
      TRACE(Authen,"client certificate issuer : "<<str);
      TRACE(Authen,"Setting dn="<<clientdn<<" roles="<<vomsroles);
      OPENSSL_free (str);
    } else {
      TRACE(Authen,"client certificate issuer : none");
      Fatal(error,"no client issuer",-1);
      if (theFD>=0) {close(theFD);theFD=-1;}
      SSLMutex.UnLock();
      return;
    }
  } else {
    TRACE(Authen,"Client does not have certificate.");
    Fatal(error,"no client certificate",-1);
    if (theFD>=0) {close(theFD);theFD=-1;}
    SSLMutex.UnLock();
    return;
  }

  // receive client proxy - if he send's one
  err = SSL_read(ssl,proxyBuff, sizeof(proxyBuff));
  if (err>0) {
    TRACE(Authen,"Recevied proxy buffer with " << err << " bytes");
    Entity.endorsements = proxyBuff;
    err = SSL_write(ssl,"OK\n",3);
    if (err!=3) {
      Fatal(error,"could not send end of handshake OK",-1);
      if (theFD>=0) {close(theFD);theFD=-1;}
      SSLMutex.UnLock();
      return;
    }
  } else {
    TRACE(Authen,"Received no proxy");
  }
  
  SSL_shutdown(ssl);

  strncpy(Entity.prot,"ssl", sizeof(Entity.prot));
  Entity.name = strdup(clientdn.c_str());
  Entity.grps = strdup(vomsroles.c_str());

  XrdOucString vomsrole = vomsroles.c_str();

  if (vomsroles.length()) {
    int dp = vomsrole.find(":");
    if (dp != STR_NPOS) {
      vomsrole.assign(vomsroles,0,dp-1);
    }
    Entity.role = strdup(vomsrole.c_str());
  } else {
    Entity.role = strdup("");
  }

  TRACE(Authen,"name = |" << Entity.name << "| role = |" << Entity.role << "| grps = |"<< Entity.grps << "|");

  if (theFD>=0) {close(theFD); theFD=-1;}       
  SSLMutex.UnLock();

  return;
}

int 
XrdSecProtocolssl::GenerateSession(const SSL* ssl, unsigned char *id, unsigned int *id_len) {
  EPNAME("GenerateSession");
  unsigned int count = 0;
  do      {
    RAND_pseudo_bytes(id, *id_len);
    /* Prefix the session_id with the required prefix. NB: If our
     * prefix is too long, clip it - but there will be worse effects
     * anyway, eg. the server could only possibly create 1 session
     * ID (ie. the prefix!) so all future session negotiations will
     * fail due to conflicts. */
    memcpy(id, "xrootdssl",
	   (strlen("xrootdssl") < *id_len) ?
	   strlen("xrootdssl") : *id_len);
   TRACE(Authen,"Generated SSID **********************");
  }
  while(SSL_has_matching_session_id(ssl, id, *id_len) &&
	(++count < MAX_SESSION_ID_ATTEMPTS));
  if(count >= MAX_SESSION_ID_ATTEMPTS)
    return 0;
  return 1;
}

int 
XrdSecProtocolssl::NewSession(SSL* ssl, SSL_SESSION *session) {
  EPNAME("NewSession");
  TRACE(Authen,"Creating new Session");
  char session_id[1024];
  for (int i=0; i< session->session_id_length; i++) {
    sprintf(session_id+(i*2),"%02x",session->session_id[i]);
  }
  if (session->cipher) {
    DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Cipher: " << session->cipher->name  << "Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
  } else {
    DEBUG("Info: ("<<__FUNCTION__<<") Session Id: "<< session_id << " Verify: " << session->verify_result << " (" << X509_verify_cert_error_string(session->verify_result) << ")");
  }
  
  SSL_set_timeout(session, sslsessionlifetime);
}

/******************************************************************************/
/*                X r d S e c p r o t o c o l u n i x I n i t                 */
/******************************************************************************/
  
extern "C"
{
char  *XrdSecProtocolsslInit(const char     mode,
                              const char    *parms,
                              XrdOucErrInfo *erp)
{
  pthread_t tid;
  int rc;

  EPNAME("ProtocolsslInit");
  // Initiate error logging and tracing
  XrdSecProtocolssl::eDest.logger(&XrdSecProtocolssl::Logger);
  
  GRSTerrorLogFunc = &MyGRSTerrorLogFunc;

  // create the tracer
  SSLxTrace = new XrdOucTrace(&XrdSecProtocolssl::eDest);
  // read the configuration options
  if (mode == 's') {
    XrdSecProtocolssl::isServer = 1;
    if (parms){
      // Duplicate the parms
      char parmbuff[1024];
      strlcpy(parmbuff, parms, sizeof(parmbuff));
      //
      // The tokenizer
      XrdOucTokenizer inParms(parmbuff);
      char *op;

      while (inParms.GetLine()) { 
	while ((op = inParms.GetToken())) {
	  if (!strncmp(op, "-d:",3)) {
	    XrdSecProtocolssl::debug = atoi(op+3);
	  } else if (!strncmp(op, "-cadir:",7)) {
	    XrdSecProtocolssl::sslcadir = strdup(op+7);
	  } else if (!strncmp(op, "-vomsdir:",6)) {
	    XrdSecProtocolssl::sslvomsdir = strdup(op+6);
	  } else if (!strncmp(op, "-cert:",6)) {
	    XrdSecProtocolssl::sslcertfile = strdup(op+6);
	  } else if (!strncmp(op, "-key:",5)) {
	    XrdSecProtocolssl::sslkeyfile = strdup(op+5);
	  } else if (!strncmp(op, "-ca:",4)) {
	    XrdSecProtocolssl::verifydepth = atoi(op+4);
	  } else if (!strncmp(op, "-t:",3)) {
	    XrdSecProtocolssl::sslsessionlifetime = atoi(op+3);
	  } else if (!strncmp(op, "-export:",8)) {
	    XrdSecProtocolssl::sslproxyexportdir = strdup(op+8);
	  }
	}
      }
    }
  }

  if (mode == 'c') {
    // default the cert/key file to the standard proxy locations
    char proxyfile[1024];
    sprintf(proxyfile,"/tmp/x509up_u%d",(int)geteuid());

    XrdSecProtocolssl::sslcertfile = strdup(proxyfile);
    XrdSecProtocolssl::sslkeyfile  = strdup(proxyfile);

    char *nossl = getenv("XrdSecNoSSL");
    if (nossl) {
      erp->setErrInfo(ENOENT,"");
      return 0;
    }

    char *cenv = getenv("XrdSecDEBUG");
    // debug
    if (cenv)
      if (cenv[0] >= 49 && cenv[0] <= 51) XrdSecProtocolssl::debug = atoi(cenv);
    
    // directory with CA certificates
    cenv = getenv("XrdSecSSLCADIR");
    if (cenv)
      XrdSecProtocolssl::sslcadir = strdup(cenv);
    else {
      // accept X509_CERT_DIR 
      cenv = getenv("X509_CERT_DIR");
      if (cenv) {
	XrdSecProtocolssl::sslcadir = strdup(cenv);
      }
    }
    // directory with VOMS certificates
    cenv = getenv("XrdSecSSLVOMSDIR");
    if (cenv)
      XrdSecProtocolssl::sslvomsdir = strdup(cenv);
    
    // file with user cert
    cenv = getenv("XrdSecSSLUSERCERT");
    if (cenv)
      XrdSecProtocolssl::sslcertfile = strdup(cenv);  
    
    // file with user key
    cenv = getenv("XrdSecSSLUSERKEY");
    if (cenv)
      XrdSecProtocolssl::sslkeyfile = strdup(cenv);
    
    // verify depth
    cenv = getenv("XrdSecSSLVERIFYDEPTH");
    if (cenv)
      XrdSecProtocolssl::verifydepth = atoi(cenv);
    
    // proxy forwarding
    cenv = getenv("XrdSecSSLPROXYFORWARD");
    if (cenv)
      XrdSecProtocolssl::forwardProxy = atoi(cenv);

    // ssl session reuse
    cenv = getenv("XrdSecSSLSESSION");
    if (cenv)
      XrdSecProtocolssl::allowSessions = atoi(cenv);
  }

  if (XrdSecProtocolssl::debug >= 3) {
    SSLxTrace->What = TRACE_ALL;
  } else if (XrdSecProtocolssl::debug == 2 ) {
    SSLxTrace->What |= TRACE_Authen;
    SSLxTrace->What |= TRACE_Debug;
  } else if (XrdSecProtocolssl::debug == 1) {
    SSLxTrace->What = TRACE_Debug;
  }

  if (XrdSecProtocolssl::isServer) {
    TRACE(Authen,"====> debug         = " << XrdSecProtocolssl::debug);
    TRACE(Authen,"====> cadir         = " << XrdSecProtocolssl::sslcadir);
    TRACE(Authen,"====> keyfile       = " << XrdSecProtocolssl::sslkeyfile);
    TRACE(Authen,"====> certfile      = " << XrdSecProtocolssl::sslcertfile);
    TRACE(Authen,"====> verify depth  = " << XrdSecProtocolssl::verifydepth);
    TRACE(Authen,"====> sess.lifetime = " << XrdSecProtocolssl::sslsessionlifetime);
  } else {
    if (XrdSecProtocolssl::debug) {
      TRACE(Authen,"====> debug         = " << XrdSecProtocolssl::debug);
      TRACE(Authen,"====> cadir         = " << XrdSecProtocolssl::sslcadir);
      TRACE(Authen,"====> keyfile       = " << XrdSecProtocolssl::sslkeyfile);
      TRACE(Authen,"====> certfile      = " << XrdSecProtocolssl::sslcertfile);
      TRACE(Authen,"====> verify depth  = " << XrdSecProtocolssl::verifydepth);
    }
  }

  if (XrdSecProtocolssl::isServer) {
    SSL_METHOD *meth;
    // initialize openssl until the context is created 
    SSL_load_error_strings();
    SSLeay_add_ssl_algorithms();
    
    meth = (SSL_METHOD*)SSLv3_server_method();
    XrdSecProtocolssl::ctx = SSL_CTX_new (meth);
    if (!XrdSecProtocolssl::ctx) {
      ERR_print_errors_fp(stderr);
      return 0;
    }
      
    if (SSL_CTX_use_certificate_file(XrdSecProtocolssl::ctx, XrdSecProtocolssl::sslcertfile, SSL_FILETYPE_PEM) <= 0) {
      ERR_print_errors_fp(stderr);
      return 0;
    }
    
    if (SSL_CTX_use_PrivateKey_file(XrdSecProtocolssl::ctx,XrdSecProtocolssl:: sslkeyfile, SSL_FILETYPE_PEM) <= 0) {
      ERR_print_errors_fp(stderr);
      return 0;
    }
      
    if (!SSL_CTX_check_private_key(XrdSecProtocolssl::ctx)) {
      fprintf(stderr,"Private key does not match the certificate public key\n");
      return 0;
    }
    
    SSL_CTX_load_verify_locations(XrdSecProtocolssl::ctx, NULL,XrdSecProtocolssl::sslcadir);  
    
    // create the store
    if (!XrdSecProtocolssl::store) {
      DEBUG("Created SSL CRL store: " << XrdSecProtocolssl::store);
      XrdSecProtocolssl::store = SSL_X509_STORE_create(NULL,XrdSecProtocolssl::sslcadir);
      X509_STORE_set_flags(XrdSecProtocolssl::store,0);
      XrdSecProtocolssl::storeLoadTime = time(NULL);
    }
    
    XrdSecProtocolssl::ctx->verify_mode = SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
    grst_cadir   = XrdSecProtocolssl::sslcadir;
    grst_vomsdir = XrdSecProtocolssl::sslvomsdir;
    grst_depth   = XrdSecProtocolssl::verifydepth;
    
    SSL_CTX_set_cert_verify_callback(XrdSecProtocolssl::ctx,
				     GRST_verify_cert_wrapper,
				     (void *) NULL);
    
    SSL_CTX_set_verify(XrdSecProtocolssl::ctx, XrdSecProtocolssl::ctx->verify_mode,GRST_callback_SSLVerify_wrapper);
    SSL_CTX_set_verify_depth(XrdSecProtocolssl::ctx, XrdSecProtocolssl::verifydepth + 1);

    if(!SSL_CTX_set_generate_session_id(XrdSecProtocolssl::ctx, XrdSecProtocolssl::GenerateSession)) {
      TRACE(Authen,"Cannot set session generator");
      return 0;
    }

    SSL_CTX_set_quiet_shutdown(XrdSecProtocolssl::ctx,1);
    SSL_CTX_set_options(XrdSecProtocolssl::ctx,  SSL_OP_ALL | SSL_OP_NO_SSLv2);
    SSL_CTX_set_session_cache_mode(XrdSecProtocolssl::ctx, SSL_SESS_CACHE_BOTH | SSL_SESS_CACHE_NO_AUTO_CLEAR );
    SSL_CTX_set_session_id_context(XrdSecProtocolssl::ctx,(const unsigned char*) XrdSecProtocolssl::SessionIdContext,
				   strlen(XrdSecProtocolssl::SessionIdContext));
    SSL_CTX_sess_set_new_cb(XrdSecProtocolssl::ctx, XrdSecProtocolssl::NewSession);
  }
  return (char *)"";
}
}

/******************************************************************************/
/*              X r d S e c P r o t o c o l u n i x O b j e c t               */
/******************************************************************************/
  
extern "C"
{
XrdSecProtocol *XrdSecProtocolsslObject(const char              mode,
                                         const char             *hostname,
                                         const struct sockaddr  &netaddr,
                                         const char             *parms,
                                               XrdOucErrInfo    *erp)
{
   XrdSecProtocolssl *prot;

// Return a new protocol object
//
   if (!(prot = new XrdSecProtocolssl(hostname, &netaddr)))
      {const char *msg = "Secssl: Insufficient memory for protocol.";
       if (erp) erp->setErrInfo(ENOMEM, msg);
          else cerr <<msg <<endl;
       return (XrdSecProtocol *)0;
      }

// All done
//
   return prot;
}
}

char*  XrdSecProtocolssl::sslcadir=(char*)"/etc/grid-security/certificates";
char*  XrdSecProtocolssl::sslvomsdir=(char*)"/etc/grid-security/vomsdir";
int    XrdSecProtocolssl::verifydepth=10;
int    XrdSecProtocolssl::verifyindex=0;
char*  XrdSecProtocolssl::sslkeyfile=(char*)"/etc/grid-security/hostkey.pem";
char*  XrdSecProtocolssl::sslcertfile=(char*)"/etc/grid-security/hostcert.pem";
char*  XrdSecProtocolssl::sslproxyexportdir=(char*)"/tmp";
int    XrdSecProtocolssl::debug=0;
time_t XrdSecProtocolssl::sslsessionlifetime=86400;
bool   XrdSecProtocolssl::isServer=0;
bool   XrdSecProtocolssl::forwardProxy=0;
bool   XrdSecProtocolssl::allowSessions=0;
char*  XrdSecProtocolssl::SessionIdContext = (char*)"xrootdssl"; 

X509_STORE*    XrdSecProtocolssl::store=0;
X509_LOOKUP*   XrdSecProtocolssl::lookup=0;
SSL_CTX*       XrdSecProtocolssl::ctx=0;
XrdSysError    XrdSecProtocolssl::eDest(0, "secssl_");
XrdSysLogger   XrdSecProtocolssl::Logger;
time_t         XrdSecProtocolssl::storeLoadTime;

XrdSysMutex XrdSecsslSessionLock::sessionmutex;
